package net.psammead.util;

import java.io.*;
import java.net.*;

/** IO functions */
public final class IOUtil {
	private static final Logger	log	= new Logger(IOUtil.class);
		
	/** fully static utility class, shall not be instantiated */
	private IOUtil() {}
	
	//-------------------------------------------------------------------------
	
	/** read a String from an URL with a given encoding  */
	public static String readStringFromURL(URL url, String charSet) throws IOException {
		return readStringFromStream(url.openConnection().getInputStream(), charSet);
	}
	
	/** read a String from a File */
	public static String readStringFromFile(File file, String charSet) throws IOException {
		return readStringFromStream(new FileInputStream(file), charSet);
	}
	
	/** read a String from an InputStream with a given encoding and close the Stream */
	public static String readStringFromStream(InputStream stream, String charSet) throws IOException {
		char[]			chars	= new char[16384];
		StringBuilder	out		= new StringBuilder();
		Reader			in		= null;
		try {
			in	= new InputStreamReader(stream, charSet);	 
			for (;;) {
				int	len	= in.read(chars);
				if (len < 0)	break; 
				out.append(chars, 0, len);
			}
			return out.toString();
		}
		finally {
            closeSilent(in);
		}
	}
	
	//-------------------------------------------------------------------------
	
	/** write a String into a File with a given encoding */
	public static void writeStringToFile(File target, String text, String charSet) throws IOException {
		writeStringToStream(new FileOutputStream(target), text, charSet);
	}
	
	/** write a String into an OutputStream with a given encoding and close the Stream */
	public static void writeStringToStream(OutputStream stream, String text, String charSet) throws IOException {
		Writer	out	= null;
		try {
			out	= new OutputStreamWriter(stream, charSet);
			out.write(text);
		}
		finally {
            closeSilent(out);
		}
	}
	
	//-------------------------------------------------------------------------
	
	/** writes a byte array into a Stream and clos the Stream  */
	public static  void writeBytesToStream(byte[] bytes, OutputStream os) throws IOException {
		try { os.write(bytes); }
		finally { closeSilent(os); }
	}
	
	/** read a byte array from a Stream and close the Stream */
	public static  byte[] readBytesFromStream(InputStream is) throws IOException {
		ByteArrayOutputStream bs = new ByteArrayOutputStream();
		try { copyStream(is, bs); }
		finally { closeSilent(is); }
		return bs.toByteArray();
	}
	
	//-------------------------------------------------------------------------
	
	/** copy the contents of an InputStream into an OutputStream */
    public static void copyStream(InputStream is, OutputStream os) throws IOException {
    	byte[]	buffer	= new byte[16384];
    	for (;;) {
    		int	count	= is.read(buffer, 0, buffer.length);
    		if (count == -1)	break;
    		os.write(buffer, 0, count);
    	}
    }
    
	//-------------------------------------------------------------------------
    
    /** close an InputStream ignoring any problems */
    public static void closeSilent(InputStream st) {
        if (st == null) return;
        try { st.close(); }
        catch (IOException e) { log.error("cannot close", e); }
    }
    
    /** close an OutputStream ignoring any problems */
    public static void closeSilent(OutputStream st) {
        if (st == null) return;
        try { st.close(); }
        catch (IOException e) { log.error("cannot close", e); }
    }
    
    /** close a Reader ignoring any problems */
    public static void closeSilent(Reader st) {
        if (st == null) return;
        try { st.close(); }
        catch (IOException e) { log.error("cannot close", e); }
    }
    
    /** close a Writer ignoring any problems */
    public static void closeSilent(Writer st) {
        if (st == null) return;
        try { st.close(); }
        catch (IOException e) { log.error("cannot close", e); }
    }
}
